/************************************************************************  
*   PROGRAMA DE INICIAO CIENTFICA PIBIC/CNPq                         *
*   PROJETO: Desenvolvimento de um simulador de algoritmos qunticos    *
*            utilizando a computao convencional                       *
* --------------------------------------------------------------------- *
*      Bolsista: Anderson de Rezende Rocha - DCC/UFLA                   *
*                   # undersun@comp.ufla.br         (1)                 *
*                   # phiberoptic_br@yahoo.com.br   (2)                 *
* --------------------------------------------------------------------- *
*  Orientadores: Antnio Maria Pereira de Resende                       *
*                   # tonio@comp.ufla.br                                *
*                Antonio Tavares Costa Junior                           *
*                   # antc@stout.ufla.br                                *
*************************************************************************/

import java.util.Random;

public class Registradores
{
    /* METODOS DISPONIVEIS
        + Registradores(int nroPosicoes);
        + Registradores();
        + Registradores(const Registradores r1);
        + int medicao();
        + void espiar(bool chave);
        + setEstado(Complexos nEstado[]);
        + void setMedia(int n);
        + void normalizar();
        + Complexos getProbabilidade(int idEstado);
		+ Complexos[] getEstados();
		+ tamanho();
    */ 
    
    // Guarda o numero de posies ou o tamanho do registrador
	private int nroPosicoes;
	// So os estados qunticos propriamente ditos
	private Complexos estados[];

	public Complexos[] getEstados()
	{
		return estados;
	}
	
	// Construtor padro de um registrador quntico
	// Serve apenas para fazer uma referncia dado que no
	// existe um registrador real propriamente dito
	public Registradores()
    {
        nroPosicoes = 0;
        estados = null;
    }

    // Este construtor realmente aloca um registrador quntico
	// alm de manter uma referncia para o mesmo
	public Registradores(int nroPosicoes)
    {
        this.nroPosicoes = nroPosicoes;
        estados = new Complexos[(int)Math.pow(2, nroPosicoes)];
		
		for (int i = 0; i < estados.length; i++)
			estados[i] = new Complexos();
    }

    // Copia um registrador para outro
    public void copiarRegistrador(Registradores antigo)
    {
        nroPosicoes = antigo.tamanho();
        int comprimentoReg = (int)Math.pow(2, nroPosicoes);
        estados = new Complexos[comprimentoReg];

		for(int i = 0; i < estados.length; i++)
            estados[i] = antigo.estados[i];
    }

	// Retorna a probabilidade de um dado estado.  usado,
	// principalmente na transformao discreta de Fourier.
	//  interessante observar que num computador quntico
	// real esta operao no seria possvel devido  decoerncia
	// quntica. Um possvel sada seria simplesmente construir
	// um circuito quntico especializado em transformao de 
	// e ento pegar apenas o resultado til e no os clculos
	// intermedirios. MAS ISTO FOGE AO ESCOPO DESTE TRABALHO
	public Complexos getProbabilidade(int idEstado)
    {
        if (idEstado >= estados.length)
		{
            System.out.println("Tentando acessar posicao de memoria invalida");
			return null;
		}
        else
            return estados[idEstado];
    }
    
	// Normaliza a probabilidade de amplitude, garante que a soma dos quadrados
	// de todos os reais e imaginrios seja sempre 1.
	public void normalizar() 
	{
		double b;
		double f, g;
		b = 0;
		for (int i = 0; i < Math.pow(2, nroPosicoes); i++) 
			b += Math.pow(estados[i].getReal(), 2) + Math.pow(estados[i].getImaginaria(), 2);
	
		b = Math.pow(b, -0.5);
	  
		for (int i = 0; i < Math.pow(2, nroPosicoes) ; i++) 
		{
			f = estados[i].getReal() * b;
			g = estados[i].getImaginaria() * b;
			estados[i].set(f, g);
		}
	}
    
	// Apenas retorna o tamanho do registrador
	public int tamanho() { return nroPosicoes; } 

	// Faz a medicao de um estado e retorna o valor DECIMAL medido.
	// Alm disso faz uma espcie de COLALPSO no estado quntico assim
	// que a medio  feita
	public int medicao()
    {
		boolean feito = false;
		int valorMedido = -1;

		double a, b;

		Random rand = new Random();
        double rand1 = rand.nextDouble();
		a = b = 0;
		for (int i = 0; i < Math.pow(2, nroPosicoes); i++)
		{
			b += Math.pow(estados[i].getReal(), 2) + Math.pow(estados[i].getImaginaria(), 2);
			if (b > rand1 && rand1 > a)
			{
				for (int j =0; j < Math.pow(2, nroPosicoes); j++)
					estados[j].set(0,0);
	
				estados[i].set(1,0);
				valorMedido = i;
				feito = true;
			}
			
			a += Math.pow(estados[i].getReal(), 2) + Math.pow(estados[i].getImaginaria(), 2);
		}

		return valorMedido;
    }


	// Imprime informaes a respeito do registrador. NAO existe realizao
	// fsica deste mtodo dado que no  possvel observar um estado
	// quntico sem colaps-lo.
	public void espiar(boolean chave)
	{
	    for (int i = 0 ; i < Math.pow(2, nroPosicoes) ; i++) 
		{
			if (chave || 
				Math.abs(estados[i].getReal()) > Math.pow(10,-15) ||
	            Math.abs(estados[i].getImaginaria()) > Math.pow(10,-15)) 
		    {
			    System.out.println("Estado " + i + " tem amplitude de probabilidade " 
				     + estados[i].getReal() + "," + estados[i].getImaginaria() + "i");
	        }
		}
	}

	// Seta o estado do registrador a partir de um vetor de estados
	public void setEstado(Complexos nEstado[])
	{
        //		for (int i = 0; i < Math.pow(2, nroPosicoes); i++)
        //			estados[i].set(nEstado[i].getReal(), nEstado[i].getImaginaria());
		for (int i = 0; i < nEstado.length; i++)
			estados[i].set(nEstado[i].getReal(), nEstado[i].getImaginaria());
	}

	public void setSequencia(Complexos nEstado[])
	{
		for (int i = 0; i < nEstado.length; i++)
			estados[i].set(nEstado[i].getReal(), nEstado[i].getImaginaria());
	}

	// Seta os estados para uma superposio igual de inteiros. O coeficiente
	// beta  sempre zero
	public void setMedia(int n)
	{
		if (n >= Math.pow(2, nroPosicoes)) 
			System.out.println("Erro inicializando registrador quantico invalido");
		else
		{
			double prob;
			prob = Math.pow(n, -0.5);
	
			for (int i = 0; i <= n; i++) 
				estados[i].set(prob, 0);
		}
	}
}
